@@ -17,6 +17,16 @@ let eq_cunit x y =
   | Signature s1, Signature s2 -> Hashtbl.hash s1 == Hashtbl.hash s2 (* BUG *)
   | _ -> false
 
+[%%if coq = "8.20" || coq = "9.0"]
+let my_filtered_open = Libobject.simple_open
+let my_simple_open ?cat f = my_filtered_open ?cat (fun i v -> if Int.equal i 1 then f v)
+let is_in_section = Lib.is_in_section
+[%%else]
+let my_filtered_open = Libobject.filtered_open
+let my_simple_open = Libobject.simple_open
+let is_in_section = Global.is_in_section
+[%%endif]
+
 
 module SLMap = Map.Make(struct type t = qualified_name let compare = compare_qualified_name end)
 module SLSet = Set.Make(struct type t = qualified_name let compare = compare_qualified_name end)
@@ -207,14 +217,14 @@ module type Programs = sig
   val declare_file : program_name -> unit
   val get_nature : qualified_name -> nature
 
-  val init_program : program_name -> src list -> unit
+  val init_program : program_name -> loc:Loc.t -> src list -> unit
   val init_db : program_name -> loc:Loc.t -> (API.Ast.Loc.t * string) -> unit
   val init_file : program_name -> Digest.t * API.Compile.scoped_program -> unit
   val header_of_db : qualified_name -> cunit list
   val ast_of_file : qualified_name -> Digest.t * API.Compile.scoped_program
 
 
-  val accumulate : qualified_name -> src list -> unit
+  val accumulate : loc:Loc.t -> qualified_name -> src list -> unit
   val accumulate_to_db : qualified_name -> cunit list -> Names.Id.t list -> scope:Rocq_elpi_utils.clause_scope -> unit
   val load_command : loc:Loc.t -> string -> unit
   val load_tactic : loc:Loc.t -> string -> unit
@@ -356,7 +366,7 @@ let in_source : Names.Id.t -> string option * EC.compilation_unit * EC.flags ->
     subst_function =(fun _ -> CErrors.user_err Pp.(str"elpi: No functors"));
     load_function  = import;
     cache_function  = cache;
-    open_function = simple_open import;
+    open_function = my_filtered_open import;
     discharge_function = (fun o -> Some o);
     classify_function = (fun _ -> Keep);
   }
@@ -505,7 +515,7 @@ let get ?(fail_if_not_exists=false) p =
     ~subst:(Some (fun _ -> CErrors.user_err Pp.(str"elpi: No functors yet")))
   
   
-  let init_program name ul =
+  let init_program name ~loc:_ ul =
     ul |> List.iter (fun u ->
       let obj = in_program (name, u, Initialization) in
       Lib.add_leaf obj)
@@ -550,8 +560,6 @@ let get ?(fail_if_not_exists=false) p =
       if Int.equal i 1 ||
         (s.scope = Rocq_elpi_utils.Global && is_inside_current_library kn) ||
         s.scope = Rocq_elpi_utils.SuperGlobal then cache o in
-    let import i (_,s as o) =
-      if Int.equal i 1 then cache o in
     declare_named_object @@ { (default_object "ELPI-DB") with
       classify_function = (fun { scope; program; _ } ->
         match scope with
@@ -562,9 +570,9 @@ let get ?(fail_if_not_exists=false) p =
       load_function  = load;
       cache_function  = cache;
       subst_function = (fun (_,o) -> o);
-      open_function = simple_open import;
+      open_function = my_simple_open cache;
       discharge_function = (fun (({ scope; program; vars; } as o)) ->
-        if scope = Rocq_elpi_utils.Local || (List.exists (fun x -> Lib.is_in_section (Names.GlobRef.VarRef x)) vars) then None
+        if scope = Rocq_elpi_utils.Local || (List.exists (fun x -> is_in_section (Names.GlobRef.VarRef x)) vars) then None
         else Some o);
     }
   
@@ -573,14 +581,12 @@ let get ?(fail_if_not_exists=false) p =
       let cache ((_,kn),((_,name),p)) =
         file_name_src := SLMap.add name p !file_name_src in
       let load i ((_,kn),_ as o) = cache o in
-      let import i (_,s as o) =
-        if Int.equal i 1 then cache o in
       declare_named_object @@ { (default_object "ELPI-FILE") with
         classify_function = (fun _ -> Keep);
         load_function  = load;
         cache_function  = cache;
         subst_function = (fun (_,o) -> o);
-        open_function = simple_open import;
+        open_function = my_simple_open cache;
         discharge_function = (fun o -> Some o);
       }
 
@@ -659,11 +665,11 @@ let get ?(fail_if_not_exists=false) p =
     | None -> CErrors.user_err Pp.(str "Elpi TacticTemplate was not called")
     | Some ast -> ast
   
-  let init_program (loc,qualid) (init : src list) =
+  let init_program (loc,qualid) ~loc (init : src list) =
     if stage = Summary.Stage.Interp && Global.sections_are_opened () then
       CErrors.user_err Pp.(str "Program/Tactic/Db cannot be declared inside sections")
     else
-      init_program qualid init
+      init_program qualid ~loc init
   
   let init_db (_,qualid) ~loc header =
     if stage = Summary.Stage.Interp && Global.sections_are_opened () then
@@ -922,8 +928,13 @@ let resolve_file_path ~must_exist ~allow_absolute ~only_elpi file =
         CErrors.user_err (Pp.str msg)
   in
   (* Lookup the directory path in Coq's state. *)
+  let mk_file_path p = Filename.concat (Loadpath.physical p) relpath in
+  let loadpath = Loadpath.find_with_logical_path logpath in
   let loadpath =
-    match Loadpath.find_with_logical_path logpath with
+    if not must_exist then loadpath
+    else List.filter (fun p -> Sys.file_exists (mk_file_path p)) loadpath in
+  let loadpath =
+    match loadpath with
     | p :: [] -> p
     | []      ->
         let msg =
@@ -946,7 +957,7 @@ let resolve_file_path ~must_exist ~allow_absolute ~only_elpi file =
         CErrors.user_err (Pp.str msg)
   in
   (* Assemble the file path. *)
-  let resolved = Filename.concat (Loadpath.physical loadpath) relpath in
+  let resolved = mk_file_path loadpath in
   if not must_exist || Sys.file_exists resolved then resolved else
   let msg =
     "File reference " ^ file ^ " was resolved to " ^ resolved ^
@@ -988,7 +999,11 @@ module Synterp : Programs = struct
     clauses_to_add |> List.iter (fun (dbname,units,vs,scope) ->
       accumulate_to_db dbname units vs ~scope))
   
-  
+  let accumulate ~loc name sources =
+    accumulate name (sources:src list);
+    handle_elpi_compiler_errors ~loc (fun () -> 
+      get_and_compile ~even_if_empty:false ~loc:(Loc.make_loc (0,0)) name |> ignore)
+
 end
 module Interp : Programs = struct
   include SourcesStorage(struct
@@ -1016,6 +1031,16 @@ let () = Rocq_elpi_builtins.set_accumulate_text_to_db_interp (fun ~loc n txt sco
       let u = unit_from_string ~elpi ~base ~loc (of_coq_loc loc) txt in
   accumulate_to_db n [u] [] ~scope)
 
+let accumulate ~loc name sources =
+  accumulate name (sources:src list);
+  handle_elpi_compiler_errors ~loc (fun () -> 
+    get_and_compile ~even_if_empty:false ~loc name |> ignore)
+
+let init_program name ~loc sources =
+  init_program name ~loc (sources:src list);
+  handle_elpi_compiler_errors ~loc (fun () -> 
+    get_and_compile ~even_if_empty:true ~loc (snd name) |> ignore)
+
 end
 
 let document_builtins () =
