@@ -69,15 +69,21 @@ let feedback_error loc m = Feedback.(feedback (Message(Error,loc,[],m)))
 let patch_loc_source execution_loc error_loc msg =
   match execution_loc, error_loc with
   | _, None -> msg, execution_loc
-  | _, Some elpiloc when Loc.finer (Some execution_loc) (Some elpiloc) -> msg, elpiloc
+  | _, Some elpiloc when Loc.finer (Some elpiloc) (Some execution_loc) -> msg, elpiloc
   | { Loc.fname },Some ({ Loc.fname = elpiname } as elpiloc) when fname = elpiname ->
       (* same file, far location, we complementwith a feedback *)
       feedback_error (Some elpiloc) msg;
-      Pp.(hv 0 (Loc.pr elpiloc ++ spc () ++ msg)), execution_loc    
+      Pp.(hv 0 (Loc.pr elpiloc ++ spc () ++ Loc.pr execution_loc ++ spc () ++ msg)), execution_loc    
   | _, Some elpiloc ->
       (* external file *)
       Pp.(hv 0 (Loc.pr elpiloc ++ spc () ++ msg)), execution_loc
 
+[%%if coq = "8.20" || coq = "9.0" || coq = "9.1"]
+exception ParseError = Gramlib.Grammar.Error
+[%%else]
+exception ParseError = Gramlib.Grammar.ParseError
+[%%endif]
+
 let handle_elpi_compiler_errors ~loc ?error_header f =
   let w_header m =
     match error_header with
@@ -95,7 +101,7 @@ let handle_elpi_compiler_errors ~loc ?error_header f =
     let _,info = Exninfo.capture e in
     let msg, loc = patch_loc_source loc (Some (to_coq_loc oloc)) (Pp.str msg) in
     CErrors.user_err ~info ~loc (w_header msg)
-  | Gramlib.Grammar.Error _ as e ->
+  | ParseError _ as e ->
     let _,info = Exninfo.capture e in
     let cloc = Loc.get_loc info in
     let msg = CErrors.print_no_report e in
@@ -133,19 +139,18 @@ let () = API.Setup.set_anomaly (fun ?loc s -> err ?loc Pp.(str s))
 let () = API.Setup.set_type_error (fun ?loc s -> err ?loc Pp.(str s))
 let warn = CWarnings.create ~name:"elpi.runtime" ~category:elpi_cat (fun x -> x)
 let warn_impl = CWarnings.create ~name:"elpi.implication-precedence" ~category:elpi_cat (fun x -> x)
-let warn_impl_rex = Re.Str.regexp_string "The standard λProlog infix operator for implication"
 let warn_linear = CWarnings.create ~name:"elpi.linear-variable" ~category:elpi_tc_cat (fun x -> x)
-let warn_linear_rex = Re.Str.regexp ".*is linear:.*discard.*fresh variable"
 let warn_missing_types = CWarnings.create ~name:"elpi.missing-types" ~category:elpi_tc_cat (fun x -> x)
-let warn_missing_types_rex = Re.Str.regexp_string "Undeclared globals"
+let warn_flex_clause = CWarnings.create ~name:"elpi.flex-clause" ~category:elpi_tc_cat (fun x -> x)
 
-let () = API.Setup.set_warn (fun ?loc x ->
+let () = API.Setup.set_warn (fun ?loc ~id x ->
   let loc = Option.map to_coq_loc loc in
   let msg = Pp.(hv 0 (Option.cata (fun loc -> Loc.pr loc ++ spc ()) (mt()) loc ++ str x)) in
-  if Re.Str.string_match warn_impl_rex x 0 then warn_impl ?loc msg
-  else if Re.Str.string_match warn_linear_rex x 0 then warn_linear ?loc msg
-  else if Re.Str.string_match warn_missing_types_rex x 0 then warn_missing_types ?loc msg
-  else warn ?loc msg)
+  match id with
+  | API.Setup.ImplicationPrecedence -> warn_impl ?loc msg
+  | LinearVariable -> warn_linear ?loc msg
+  | UndeclaredGlobal -> warn_missing_types ?loc msg
+  | FlexClause -> warn_flex_clause ?loc msg)
 let () = API.Setup.set_std_formatter (Format.make_formatter feedback_fmt_write feedback_fmt_flush)
 let () = API.Setup.set_err_formatter (Format.make_formatter feedback_fmt_write feedback_fmt_flush)
 let nYI s = err Pp.(str "Not Yet Implemented: " ++ str s)
@@ -287,8 +292,12 @@ let (projectionc, projection) : Names.Projection.t Elpi.API.RawOpaqueData.cdata
       name = "projection";
       doc = "";
       pp = (fun fmt i -> Format.fprintf fmt "%s" (Names.Projection.to_string i));
-      compare = Names.Projection.CanOrd.compare;
-      hash = Names.Projection.CanOrd.hash;
+      compare =
+        (let open Names.Projection in
+         fun x y -> Repr.CanOrd.compare (repr x) (repr y));
+      hash =
+        (let open Names.Projection in
+         fun x -> Repr.CanOrd.hash (repr x));
       hconsed = false;
       constants = [];
     }
@@ -510,7 +519,7 @@ let detype_primitive_string = function
   | Constr.String s -> DAst.make @@ Glob_term.GString s
   | _ -> assert false
 
-[%%if coq = "8.19" || coq = "8.20" ]
+[%%if coq = "8.20" ]
 let fresh (names, e) sigma name ty =
   let open EConstr in
   let open Names.Name in
@@ -753,7 +762,7 @@ let detype ?(keepunivs = false) env sigma t =
   let x = aux (names_of_env env, env) t in
   x
 
-[%%if coq = "8.19" || coq = "8.20" ]
+[%%if coq = "8.20" ]
 let detype_closed_glob env sigma closure =
   let gbody = Detyping.detype_closed_glob Names.Id.Set.empty env sigma closure in
   fix_detype gbody
@@ -763,6 +772,17 @@ let detype_closed_glob env sigma closure =
   fix_detype gbody
 [%%endif]
 
+[%%if coq = "8.20" || coq = "9.0" || coq = "9.1"]
+let detype_to_pattern env sigma c =
+  let c = detype env sigma c in
+  Patternops.pattern_of_glob_constr env c
+[%%else]
+let detype_to_pattern env sigma c =
+  let c = detype env sigma c in
+  let ids, p = Patternops.pattern_of_glob_constr env c in
+  ids, Patternops.interp_pattern env sigma Glob_ops.empty_lvar p
+[%%endif]
+
 type qualified_name = string list
 
 let compare_qualified_name = Stdlib.compare
